/*
*	trans: string translator
*	ver. 1.1
*	-Kenar- (Kenji Arisawa)
*	Email: arisawa@aichi-u.ac.jp
*/

#define DBG if(0)

#ifdef PLAN9
#include <u.h>
#include <libc.h>
#include <bio.h>
Biobuf bout;
#else
#include <plan9.h>
#include <stdio.h>
#include <stdlib.h>	/* for malloc */
#include "ubio.h"
#include "misc.h"
#define UNIX	1
int open(char*,int);
#endif

#include "trans.h"

extern Biobuf bout;
extern int binmode;

void
usage(void)
{
	fprint(2,
	"usage: trans [-b] rulefile [file ....]\n"
	"name:\n"
	"  translate strings using rules in the rulefile\n"
	"options:\n"
	" -b: binary mode\n"
	);
	exits("usage");
}


/* 	we have two types of translate()
*	Btranslate() and stranslate()
*	Btranslate() is better for the tool that handle stream
*	stranslate() is designed for string conversion.
*	stranslate() is only for debugging purpose.
*	we need the codes in stranslate() for u9fs */

#define translate stranslate

void
doit(int argc, char *argv[])
{
	int fd;
	TrTab trtab;

	ARGBEGIN{
	case 'b': binmode++; break;
	default: usage();
	}ARGEND
	if(*argv == nil)
		usage();

	Binit(&bout, 1, OWRITE);
	DBG print("# Binit done\n");

	tabread(&trtab, *argv++);
	DBG print("# tabread done\n");

	if(*argv == nil){
		translate(0,&trtab);
		exits(nil);
	}

	while(*argv){
		fd = open(*argv++,OREAD);
		translate(fd,&trtab);
	}
	tabfree(&trtab);
}

#ifdef PLAN9
void
main(int argc, char *argv[])
{
	doit(argc, argv);
	exits(nil);
}
#else
int
main(int argc, char *argv[])
{
	doit(argc, argv);
	return 0;
}
#endif
