/*
*	trans: string translator
*	ver. 1.0
*	-Kenar- (Kenji Arisawa)
*	Email: arisawa@aichi-u.ac.jp
*/

/*
*	PLAN9:	Plan9 and P9P
*	UNIX:	UNIX and P9P
*/

#ifdef PLAN9
/* include Plan9 port */
#include <u.h>
#include <libc.h>
#include <bio.h>
#else
#include <plan9.h>
#include <stdlib.h>	/* for malloc */
#include "ubio.h"
/*	print() (in print.c) is buggy, so I replaced by printf() in stdio.h .
* look also "Segmentation fault" in this file. */
#define print printf
#define UNIX	1
int open(char*,int);
#endif

#include "trans.h"

#define	DBG	if(0)
#define	DBG0 if(0)
/*	note: if(0) { ... }
*	if(0) strips the object codes from *.$O.
*	you need not remove debugging codes from the source to save object code size.
*/

#ifdef UNIX
#define stime s_time
#endif


#define ERR (-1)

int _ungetcount;
uchar _ungetbuf[512];

enum{
	/* some rule file formats */
	rfmt = 0, /* raw utf-8 string, default */
	hfmt = 1, /* utf-8 in hex */
	ufmt = 2, /* unicode in hex */
};

#ifndef PLAN9

#endif



Biobuf bout;
int binmode;

List* Lnew(void *d)
{
	List *p;
	p = (List*) malloc(sizeof(List));
	p->next = 0;
	p->data = d;
	return p;
};

List* Lappend(List *p, void *d)
{
	List *q;
	q = (List*) malloc(sizeof(List));
	q->next = p->next;
	q->data = d;
	p->next = q;
	return q;
};

void Lfree(List *p)
{
	List *q;
	q = p->next;
	free(p);
	while(q != 0){
		p = q->next;
		free(q);
		q = p;
	}
}

#ifdef XXX
/*	this main() show the usage of
*	Lnew(), Lappend() and Lfree()
*	these are simple function for List	*/
void main(int argc, char *argv[])
{
	List *a, *p;

	USED(argc);
	USED(argv);

	a = Lnew("alice");
	p = Lappend(a,"bob");
	p = Lappend(p,"carol");

	for (p = a; p; p = p->next)
		print("%s\n", p->data);

	Lfree(a);
}
#endif

/*
*	u8rb:	gets the second utf-8 rune boundary
*	note that the first rune boundary is &bp[0] */
int
u8rb(char *bp, int m)
{
	char *p,*e;
	p = bp;
	e = bp + m;
	p++;
	while(p < e){
		if((*p & 0x80) == 0)
			break;
		if((*p & 0xc0) == 0xc0)
			break;
		p++;
	}
	return p - bp;
}


/*
 * Bgetch: extended Bgetc()
 */
int
Bgetch(Biobuf *bin)
{
	int c;
	if(_ungetcount > 0)
		return _ungetbuf[--_ungetcount];
	c = Bgetc(bin);
	if(c < 1)
		return ERR;
	return (unsigned)c&0xff;
}

/*
 * Bungetch: unget n chars.
 */
void
Bungetch(Biobuf *bip, char *s, int n)
{
	USED(bip); // dummy
	while(n > 0)
		_ungetbuf[_ungetcount++] = s[--n];
}


void
Btrans(Biobuf *bop, Biobuf *bip, TrTab *trtab)
/*	bop: pointer to bout
*	bip: pointer to bin
*	ttab: table of rule
*	n: number of rules
*	buf: working buf, the size must be >= max len of ttab[n].frstr	*/
{
	int m,j,k,c;
	Ttab *ttab;
	int n;
	char *buf;
	int i;

	ttab = trtab->ttab;
	buf = trtab->buf;
	n = trtab->n;
	DBG print("n:%d\n",n);

	for(;;){
Lstart:
		c = 0;
		m = 0; j = 0; k = n - 1;
		while(j < k) {
			c = Bgetch(bip);
			DBG print("|%02x:m,j,k:%d,%d,%d\n",c,m,j,k);
			if(c == ERR)
				goto Lend;
			while (j <= k && c > (uchar)ttab[j].frstr[m]) j++;
			while (j <= k && c < (uchar)ttab[k].frstr[m]) k--;
			buf[m++] = c;
			DBG print(">%02x:m,j,k:%d,%d,%d\n",c,m,j,k);
		}
		DBG print("*%02x:m,j,k:%d,%d,%d\n", c,m,j,k);
		if(j == k) {
			while(m < ttab[j].frlen) {
				c = Bgetch(bip);
				DBG print("^%02x,%02x:m,frlen:%d,%d\n",c,(uchar)ttab[j].frstr[m],m,ttab[j].frlen);
				if(c == ERR)
					goto Lend;
				buf[m] = c;
				/* NB: m is number of bytes in buffer */
				if((uchar)ttab[j].frstr[m++] != c)
					goto Lnext;
			}
			DBG print("@%02x:m,j:%d,%d\n",c,m,j);
			if(m == ttab[j].frlen) /* found */
				Bwrite(bop, ttab[j].tostr, strlen(ttab[j].tostr));
			goto Lstart;
		}
Lnext: /* flash buf and go next */
		if(binmode)
			i = 1;
		else
			i = u8rb(buf, m);
		Bwrite(bop, buf, i);
		Bungetch(bip, &buf[i], m - i);
	}
Lend:
	Bwrite(bop, buf, m);
}


/*
 * sgetch: read chars from string
 */
int
sgetch(char *s)
{
	int c;
	static char *b0;
	static char *p = nil;
	if(s == nil){
		b0 = nil;
		return 0;
	}
	if(b0 != s)
		p = b0 = s;
	if(_ungetcount > 0)
		return (uchar)_ungetbuf[--_ungetcount];
	if(*p == 0)
		return ERR;
	c = (uchar) *p++;
	return c;
}

/*
 * ungetch: unget n chars.
 */
void
ungetch(char *s, int n)
{
	while(n > 0)
		_ungetbuf[_ungetcount++] = s[--n];
}

char *
ewrite(char *wp, char *ep, char *s, int n)
/*
*	wp: working point
*	ep: end point
*	return: new working point	*/
{
	if(wp + n > ep)
		n = ep - wp;
	memcpy(wp,s,n);
	wp += n;
	return wp;
}

char *
etrans(char *result, char *ep, char *source, TrTab *trtab)
/*	result: buf to store the result
*	ep: size of result
*	source: source text
*	ttab: table of rule
*	n: number of rules
*	buf: working buf, the size must be >= max len of ttab[n].frstr
*	return: next writing point (not null terminated)	*/
{
	int m,j,k,c;
	Ttab *ttab;
	int n;
	char *buf;
	int i;
	char *wp = result;

	ttab = trtab->ttab;
	buf = trtab->buf;
	n = trtab->n;
	DBG print("n:%d\n",n);

	sgetch(nil); /* initialize input buffer */
	for(;;){
Lstart:
		c = 0;
		m = 0; j = 0; k = n - 1;
		while(j < k) {
			c = sgetch(source);
			DBG print("|%02x:m,j,k:%d,%d,%d\n",c,m,j,k);
			if(c == ERR)
				goto Lend;
			while (j <= k && c > (uchar)ttab[j].frstr[m]) j++;
			while (j <= k && c < (uchar)ttab[k].frstr[m]) k--;
			buf[m++] = c;
			DBG print(">%02x:m,j,k:%d,%d,%d\n",c,m,j,k);
		}
		DBG print("*%02x:m,j,k:%d,%d,%d\n", c,m,j,k);
		if(j == k) {
			while(m < ttab[j].frlen) {
				c = sgetch(source);
				DBG print("^%02x,%02x:m,frlen:%d,%d\n",c,(uchar)ttab[j].frstr[m],m,ttab[j].frlen);
				if(c == ERR)
					goto Lend;
				buf[m] = c;
				/* NB: m is number of bytes in buffer */
				if((uchar)ttab[j].frstr[m++] != c)
					goto Lnext;
			}
			DBG print("@%02x:m,j:%d,%d\n",c,m,j);
			if(m == ttab[j].frlen) /* found */
				wp = ewrite(wp,ep,ttab[j].tostr, strlen(ttab[j].tostr));
			goto Lstart;
		}
Lnext: /* flash buf and go next */
		if(binmode)
			i = 1;
		else
			i = u8rb(buf, m);
		wp = ewrite(wp, ep, buf, i);
		ungetch(&buf[i], m - i);
	}
Lend:
	wp = ewrite(wp,ep, buf, m);
	return wp;
}

char *
trfmt(char *str, int fmt)
/*	tr translate various formatted rule data to utf-8 raw format.
*	str: input
*	fmt: format style
*	return: converted str.
*	NB: str will be scratched */
{
	char *s,*end;
	s = str;

	if(fmt == hfmt){
		int i,ch;
		for(i = 0;;){
			ch = strtol(s,&end,16);
			str[i++] = ch;
			if(*end == 0)
				break;
			s = end+1;
		}
		str[i] = 0;
		return str;
	}
	if(fmt == ufmt){
		int i,n;
		Rune rune;
		for(i = 0;;){
			rune = strtol(s,&end,16);
			/* typedef	ushort		Rune;	*/
			n = runetochar(&str[i],&rune);
			i += n;
			if(*end == 0)
				break;
			s = end+1;
		}
		str[i] = 0;
		return str;
	}
	return str;
}

/*	man strcat(2): 
*	The comparisons are made with unsigned bytes.	*/
int
compar(Ttab *x, Ttab *y)
{
	return strcmp(x->frstr,y->frstr);
}

/*
*	tabread: read translation data from rule file
*	lines beginning with "#" are comment
*	if the first line is comment and string "unicode" or "hex" is in it,
*	these are keywords for format of the file 	*/
void
tabread(TrTab *trtabp, char *path)
{
	int n,i,m, fd;
	int fmt = 0;
	Biobuf bin;
	Ttab *ttab, *tp;
	char *p;
	List *q, *q0;
	char *args[3];
	long ch;
	char *buf;

	DBG0 print("# begin tabread\n");
	/*
	*	trans data init
	*/
	fd = open(path, OREAD);
	if(fd < 0)
		sysfatal("%s not open %r", path);
	if(Binit(&bin, fd, OREAD) < 0)
		sysfatal("error in Binit: %r");
	ch = Bgetc(&bin);
	if(ch == '#'){
		p = Brdstr(&bin, '\n',1);
		if(cistrstr(p,"unicode"))
			fmt = ufmt;
		else
		if(cistrstr(p,"hex"))
			fmt = hfmt;
	}
	else
		Bungetc(&bin);

	trtabp->list = q0 = q = Lnew(nil);
	n = 0;
	while((p = Brdstr(&bin, '\n', 1))){ /* assign = */
		if(p[0] == '#')
			continue;
		DBG0 print("%s\n",p);
		q = Lappend(q, strdup(p));
		n++;
	}
	close(fd);
	if(n == 0)
		sysfatal("# no rule");
	DBG0 print("done\n");

	ttab = calloc(n + 1, sizeof(Ttab));
	/* need room for nil */
	q0 = q0->next;

	for(i = 0, q = q0; q; q = q->next){
		m = tokenize(q->data,args, 3);
		if(m < 2)
			continue;
		tp = &ttab[i];
		tp->frstr = trfmt(args[0],fmt);
		tp->tostr = trfmt(args[1],fmt);
		tp->frlen = strlen(tp->frstr);
		if(tp->frlen == 0)
			sysfatal("# null string in the first field\n");
		DBG0 print("#%s %s\n", args[0], args[1]);
		i++;
	}

	/* we nust sort ttab by frstr */
	DBG0 print("sort...\n");
    /*
	* Plan9
	* void qsort(void *base, long nel, long width,
    *   int (*compar)(void*, void*));
	*
	* UNIX
    * void qsort(void *base, size_t nel, size_t width,
	*   int (*compar)(const void *, const void *));
	*/
	n = i; /* num of element. don't change the meaning hereafter */

#ifdef UNIX
	qsort(ttab, n, sizeof(Ttab), (int (*)(const void*, const void*))compar);
#else
	qsort(ttab, n, sizeof(Ttab), compar);
#endif
	DBG0 print("the sort result n=%d\n",n);

	/* check frstr != nil */
	DBG0 for(i = 0; i < n; i++)
		if(ttab[i].frstr == nil)
			print("## %d %s %s\n", i, ttab[i].frstr , ttab[i].tostr);
	DBG0 print("nil check OK\n");


	/* DBG */
	DBG0 for(i = 0; i < n; i++)
		print("## %s %s\n", ttab[i].frstr , ttab[i].tostr);
		// print("## %02x %s %s\n", ttab[i].frstr[0], ttab[i].frstr , ttab[i].tostr);
		// Segmentation fault if we apply print() in print.c
		// printf() in stdio.h is OK
	m = 0;
	for(i = 0; i < n; i++){
		DBG0 print("%s %s\n",ttab[i].frstr,ttab[i].tostr);
		if(ttab[i].frlen > m)
			m = ttab[i].frlen;
	}
	/* then m is max size of frlen */
	buf = malloc(m + 1); // need space for '\0' in case etrans()

	trtabp->ttab = ttab;
	trtabp->n = n;
	trtabp->buf = buf;
}


/*
*	Btranslate() shows the usage example of Btrans()
*	Btrans() is designed for stream conversion
*	stdin to stdout
*	Binary data should be OK.
*/
void
Btranslate(int fd, TrTab *trtab)
{
	Biobuf bin;
	if(Binit(&bin, fd, OREAD) < 0)
		sysfatal("input not open: %r");
	Btrans(&bout, &bin, trtab);
	Bflush(&bout);
}


/*
*	stranslate() shows the usage example of etrans()
*	etrans() is designed for string to string conversion
*	these string must be utf-8.
*/
void
stranslate(int fd, TrTab *trtab)
{
	Biobuf bin;
	char *source;
	char obuf[4098], *ep, *wp;
	ep = obuf + sizeof(obuf);
	if(Binit(&bin, fd, OREAD) < 0)
		sysfatal("input not open: %r");
	while((source = Brdstr(&bin,'\n',1))){ /* assign = */
		DBG0 print("!%s\n",source);
		wp = etrans(obuf, ep, source, trtab);
		if(wp == ep)
			fprint(2,"# output buffer full\n");
		*wp = 0; /* needed */
		Bprint(&bout,"%s\n",obuf);
	}
	Bflush(&bout);
}

void
tabfree(TrTab *tp)
{
	List *q, *q0;
	Ttab *ttab = tp->ttab;

	free(tp->buf);
	q0 = tp->list;
	for(q = q0; q; q = q->next)
		free(q->data);
	Lfree(q0);
	free(ttab);
}



#ifdef XXX

/* 	we have two types of translate()
*	Btranslate() and stranslate()
*	Btranslate() is better for the tool that handle stream
*	stranslate() is designed for string conversion.
*	stranslate() is only for debugging purpose.
*	we need the codes in stranslate() for u9fs */

#define translate Btranslate

void
main(int argc, char *argv[])
{
	int fd;
	TrTab trtab;

	ARGBEGIN{
	default: usage();
	}ARGEND

	if(*argv == nil)
		usage();

	Binit(&bout, 1, OWRITE);

	tabread(&trtab, *argv++);

	if(*argv == nil){
		translate(0,&trtab);
		exits(nil);
	}

	while(*argv){
		fd = open(*argv++,OREAD);
		translate(fd,&trtab);
	}
	tabfree(&trtab);
	exits(nil);
}

#endif
